# flake8: noqa F401
from .llama_magnitude_prune import MagnitudePruningForLlama
from .llama_random_prune import RandomPruningForLlama
from .llama_wanda_prune import WandaPruningForLlama
from .magnitude_diff_pruning import MagnitudeDiffPruningAlgorithm
