classnames = [
    "abbey",
    "airplane cabin",
    "airport terminal",
    "alley",
    "amphitheater",
    "amusement arcade",
    "amusement park",
    "anechoic chamber",
    "apartment building outdoor",
    "apse indoor",
    "aquarium",
    "aqueduct",
    "arch",
    "archive",
    "arrival gate outdoor",
    "art gallery",
    "art school",
    "art studio",
    "assembly line",
    "athletic field outdoor",
    "atrium public",
    "attic",
    "auditorium",
    "auto factory",
    "badlands",
    "badminton court indoor",
    "baggage claim",
    "bakery shop",
    "balcony exterior",
    "balcony interior",
    "ball pit",
    "ballroom",
    "bamboo forest",
    "banquet hall",
    "bar",
    "barn",
    "barndoor",
    "baseball field",
    "basement",
    "basilica",
    "basketball court outdoor",
    "bathroom",
    "batters box",
    "bayou",
    "bazaar indoor",
    "bazaar outdoor",
    "beach",
    "beauty salon",
    "bedroom",
    "berth",
    "biology laboratory",
    "bistro indoor",
    "boardwalk",
    "boat deck",
    "boathouse",
    "bookstore",
    "booth indoor",
    "botanical garden",
    "bow window indoor",
    "bow window outdoor",
    "bowling alley",
    "boxing ring",
    "brewery indoor",
    "bridge",
    "building facade",
    "bullring",
    "burial chamber",
    "bus interior",
    "butchers shop",
    "butte",
    "cabin outdoor",
    "cafeteria",
    "campsite",
    "campus",
    "canal natural",
    "canal urban",
    "candy store",
    "canyon",
    "car interior backseat",
    "car interior frontseat",
    "carrousel",
    "casino indoor",
    "castle",
    "catacomb",
    "cathedral indoor",
    "cathedral outdoor",
    "cavern indoor",
    "cemetery",
    "chalet",
    "cheese factory",
    "chemistry lab",
    "chicken coop indoor",
    "chicken coop outdoor",
    "childs room",
    "church indoor",
    "church outdoor",
    "classroom",
    "clean room",
    "cliff",
    "cloister indoor",
    "closet",
    "clothing store",
    "coast",
    "cockpit",
    "coffee shop",
    "computer room",
    "conference center",
    "conference room",
    "construction site",
    "control room",
    "control tower outdoor",
    "corn field",
    "corral",
    "corridor",
    "cottage garden",
    "courthouse",
    "courtroom",
    "courtyard",
    "covered bridge exterior",
    "creek",
    "crevasse",
    "crosswalk",
    "cubicle office",
    "dam",
    "delicatessen",
    "dentists office",
    "desert sand",
    "desert vegetation",
    "diner indoor",
    "diner outdoor",
    "dinette home",
    "dinette vehicle",
    "dining car",
    "dining room",
    "discotheque",
    "dock",
    "doorway outdoor",
    "dorm room",
    "driveway",
    "driving range outdoor",
    "drugstore",
    "electrical substation",
    "elevator door",
    "elevator interior",
    "elevator shaft",
    "engine room",
    "escalator indoor",
    "excavation",
    "factory indoor",
    "fairway",
    "fastfood restaurant",
    "field cultivated",
    "field wild",
    "fire escape",
    "fire station",
    "firing range indoor",
    "fishpond",
    "florist shop indoor",
    "food court",
    "forest broadleaf",
    "forest needleleaf",
    "forest path",
    "forest road",
    "formal garden",
    "fountain",
    "galley",
    "game room",
    "garage indoor",
    "garbage dump",
    "gas station",
    "gazebo exterior",
    "general store indoor",
    "general store outdoor",
    "gift shop",
    "golf course",
    "greenhouse indoor",
    "greenhouse outdoor",
    "gymnasium indoor",
    "hangar indoor",
    "hangar outdoor",
    "harbor",
    "hayfield",
    "heliport",
    "herb garden",
    "highway",
    "hill",
    "home office",
    "hospital",
    "hospital room",
    "hot spring",
    "hot tub outdoor",
    "hotel outdoor",
    "hotel room",
    "house",
    "hunting lodge outdoor",
    "ice cream parlor",
    "ice floe",
    "ice shelf",
    "ice skating rink indoor",
    "ice skating rink outdoor",
    "iceberg",
    "igloo",
    "industrial area",
    "inn outdoor",
    "islet",
    "jacuzzi indoor",
    "jail cell",
    "jail indoor",
    "jewelry shop",
    "kasbah",
    "kennel indoor",
    "kennel outdoor",
    "kindergarden classroom",
    "kitchen",
    "kitchenette",
    "labyrinth outdoor",
    "lake natural",
    "landfill",
    "landing deck",
    "laundromat",
    "lecture room",
    "library indoor",
    "library outdoor",
    "lido deck outdoor",
    "lift bridge",
    "lighthouse",
    "limousine interior",
    "living room",
    "lobby",
    "lock chamber",
    "locker room",
    "mansion",
    "manufactured home",
    "market indoor",
    "market outdoor",
    "marsh",
    "martial arts gym",
    "mausoleum",
    "medina",
    "moat water",
    "monastery outdoor",
    "mosque indoor",
    "mosque outdoor",
    "motel",
    "mountain",
    "mountain snowy",
    "movie theater indoor",
    "museum indoor",
    "music store",
    "music studio",
    "nuclear power plant outdoor",
    "nursery",
    "oast house",
    "observatory outdoor",
    "ocean",
    "office",
    "office building",
    "oil refinery outdoor",
    "oilrig",
    "operating room",
    "orchard",
    "outhouse outdoor",
    "pagoda",
    "palace",
    "pantry",
    "park",
    "parking garage indoor",
    "parking garage outdoor",
    "parking lot",
    "parlor",
    "pasture",
    "patio",
    "pavilion",
    "pharmacy",
    "phone booth",
    "physics laboratory",
    "picnic area",
    "pilothouse indoor",
    "planetarium outdoor",
    "playground",
    "playroom",
    "plaza",
    "podium indoor",
    "podium outdoor",
    "pond",
    "poolroom establishment",
    "poolroom home",
    "power plant outdoor",
    "promenade deck",
    "pub indoor",
    "pulpit",
    "putting green",
    "racecourse",
    "raceway",
    "raft",
    "railroad track",
    "rainforest",
    "reception",
    "recreation room",
    "residential neighborhood",
    "restaurant",
    "restaurant kitchen",
    "restaurant patio",
    "rice paddy",
    "riding arena",
    "river",
    "rock arch",
    "rope bridge",
    "ruin",
    "runway",
    "sandbar",
    "sandbox",
    "sauna",
    "schoolhouse",
    "sea cliff",
    "server room",
    "shed",
    "shoe shop",
    "shopfront",
    "shopping mall indoor",
    "shower",
    "skatepark",
    "ski lodge",
    "ski resort",
    "ski slope",
    "sky",
    "skyscraper",
    "slum",
    "snowfield",
    "squash court",
    "stable",
    "stadium baseball",
    "stadium football",
    "stage indoor",
    "staircase",
    "street",
    "subway interior",
    "subway station platform",
    "supermarket",
    "sushi bar",
    "swamp",
    "swimming pool indoor",
    "swimming pool outdoor",
    "synagogue indoor",
    "synagogue outdoor",
    "television studio",
    "temple east asia",
    "temple south asia",
    "tennis court indoor",
    "tennis court outdoor",
    "tent outdoor",
    "theater indoor procenium",
    "theater indoor seats",
    "thriftshop",
    "throne room",
    "ticket booth",
    "toll plaza",
    "topiary garden",
    "tower",
    "toyshop",
    "track outdoor",
    "train railway",
    "train station platform",
    "tree farm",
    "tree house",
    "trench",
    "underwater coral reef",
    "utility room",
    "valley",
    "van interior",
    "vegetable garden",
    "veranda",
    "veterinarians office",
    "viaduct",
    "videostore",
    "village",
    "vineyard",
    "volcano",
    "volleyball court indoor",
    "volleyball court outdoor",
    "waiting room",
    "warehouse indoor",
    "water tower",
    "waterfall block",
    "waterfall fan",
    "waterfall plunge",
    "watering hole",
    "wave",
    "wet bar",
    "wheat field",
    "wind farm",
    "windmill",
    "wine cellar barrel storage",
    "wine cellar bottle storage",
    "wrestling ring indoor",
    "yard",
    "youth hostel",
]

templates = [
    lambda c: f"a photo of a {c}.",
    lambda c: f"a photo of the {c}.",
]
